@@ -37,8 +37,13 @@ def main_action(resource_file: str, action_name: str, env_name: str, base_path:
     for _, state in ex.get_states().items():
         res_id = state.get_resource_id()
         _deps = deps.get(res_id, [])
+        env_added = False
         for dep in _deps:
             state.add_dep(dep.get("id"), dep.get("hash"))
+            if dep.get("id", "").split(":")[0] == "Env":
+                env_added = True
+        if env_added is False and env is not None:
+            state.add_dep(f"Env:{env_name}", rs.calculate_hash(env))
         store.store(state)
     for ac in core.all_actions:
         artifacts = st.get_artifacts(ac, env_name).get(ac, [])
