@@ -3,6 +3,7 @@ import shlex
 import os
 import subprocess
 import json
+import yaml
 
 import pluggy
 from hash import resource_hookimpl as hookimpl
@@ -28,6 +29,14 @@ class ResourceSpec:
     def re_action(self, action, state, env):
         pass
 
+    @hookspec
+    def depends_on_env(self):
+        pass
+
+    @hookspec
+    def get_ignores(self):
+        pass
+
 
 class Resource(object):
     def __init__(self, file) -> None:
@@ -76,6 +85,12 @@ class Resource(object):
         """
         return ["path"]
 
+    def depends_on_env(self):
+        return True
+
+    def get_ignores(self):
+        return []
+
     def raise_resource_error(self, name, e):
         if name == "build":
             raise errors.ResourceBuildError(e)
@@ -304,6 +319,10 @@ class FakeResource(Resource):
     def init(self, file):
         pass
 
+    @hookimpl
+    def depends_on_env(self):
+        return self.getSpec("depends_on_env", True)
+
     @hookimpl
     def action(self, name, state, env):
         if name not in all_actions:
@@ -379,16 +398,16 @@ class ProjectResource(Resource):
 
 class TerraformResource(Resource):
     def select_workspace(self, workspace):
-        workspace_new_command = f"terraform workspace new {workspace}"
+        workspace_new_command = f"{self.__terraform} workspace new {workspace}"
         try:
             self.execute(workspace_new_command, self.getPath())
         except Exception:
             pass
-        workspace_select_command = f"terraform workspace select {workspace}"
+        workspace_select_command = f"{self.__terraform} workspace select {workspace}"
         self.execute(workspace_select_command, self.getPath())
 
     def terraform_init(self, workspace):
-        init_command = "terraform init -upgrade -input=false -force-copy"
+        init_command = f"{self.__terraform} init -upgrade -input=false -force-copy"
         state_backend = self.getSpec("state_backend")
         if state_backend:
             kind = state_backend.get("kind", "gcs")
@@ -421,6 +440,52 @@ class TerraformResource(Resource):
         self.execute(init_command, self.getPath())
         self.select_workspace(workspace)
 
+    @hookimpl
+    def depends_on_env(self):
+        return False
+
+    def re_action(self, action, state, env):
+        if action == "publish":
+            return False
+        self.check_terraform()
+        if env is None or env == {}:
+            self.raise_resource_error(
+                "deploy", "Terraform resource requires an env with workspace argument")
+        if self.__get_version() != state.get("envs", {}).get(env.getName(), {}).get(action, {}).get("version"):
+            return True
+        if action in ["build", "test", "publish"]:
+            return False
+        workspace = env.getSpec("workspace", env.getName())
+        try:
+            self.terraform_init(workspace)
+        except subprocess.CalledProcessError as e:
+            raise errors.ResourceDeployError(
+                f"Cannot initialize terraform, {e}")
+        except subprocess.TimeoutExpired as e:
+            raise errors.ResourceDeployError(
+                f"Timeout initializing terraform, {e}")
+        build_command = f"{self.__terraform} plan -detailed-exitcode -no-color -input=false"
+        variables = self.getSpec("variables")
+        if variables:
+            vars = ""
+            for (var_name, var_value) in variables.items():
+                if type(var_value) == list:
+                    value = str(var_value).replace("\'", "\"")
+                    vars += f" -var '{var_name}={value}'"
+                else:
+                    vars += f" -var {var_name}={var_value}"
+            build_command += vars
+        try:
+            self.execute(build_command, self.getPath(),
+                         self.getSpec("plan_timeout", 300))
+        except FileNotFoundError as e:
+            raise errors.ResourceBuildError("cannot find terraform in PATH")
+        except subprocess.CalledProcessError as e:
+            if e.returncode == 2:
+                return True
+            raise errors.ResourceBuildError(e.stderr.decode("UTF-8"))
+        return False
+
     @hookimpl
     def init(self, file: str):
         """
@@ -440,7 +505,7 @@ class TerraformResource(Resource):
 
     def get_outputs(self, path) -> dict:
         try:
-            refresh_command = "terraform refresh"
+            refresh_command = f"{self.__terraform} refresh"
             variables = self.getSpec("variables")
             if variables:
                 vars = ""
@@ -455,13 +520,90 @@ class TerraformResource(Resource):
         except subprocess.CalledProcessError as e:
             raise errors.ResourceBuildError(e)
         try:
-            proc = self.execute("terraform output -json", path)
+            proc = self.execute(f"{self.__terraform} output -json", path)
             return json.loads(proc.stdout)
         except subprocess.CalledProcessError as e:
             raise errors.ResourceBuildError(e)
 
+    def __install_terraform(self, version):
+        download_link = f"https://releases.hashicorp.com/terraform/{version}/terraform_{version}_linux_amd64.zip"
+        terraform_check_sums = {
+            "1.3.9": "53048fa573effdd8f2a59b726234c6f450491fe0ded6931e9f4c6e3df6eece56",
+            "1.3.8": "9d9e7d6a9b41cef8b837af688441d4fbbd84b503d24061d078ad662441c70240",
+            "1.3.7": "b8cf184dee15dfa89713fe56085313ab23db22e17284a9a27c0999c67ce3021e",
+            "1.3.6": "bb44a4c2b0a832d49253b9034d8ccbd34f9feeb26eda71c665f6e7fa0861f49b",
+            "1.3.5": "ac28037216c3bc41de2c22724e863d883320a770056969b8d211ca8af3d477cf",
+            "1.3.4": "b24210f28191fa2a08efe69f54e3db2e87a63369ac4f5dcaf9f34dc9318eb1a8",
+            "1.3.3": "fa5cbf4274c67f2937cabf1a6544529d35d0b8b729ce814b40d0611fd26193c1",
+            "1.3.2": "6372e02a7f04bef9dac4a7a12f4580a0ad96a37b5997e80738e070be330cb11c",
+            "1.3.1": "0847b14917536600ba743a759401c45196bf89937b51dd863152137f32791899",
+            "1.3.0": "380ca822883176af928c80e5771d1c0ac9d69b13c6d746e6202482aedde7d457",
+        }
+        terraform_check_sum = self.getSpec("terraform_check_sum")
+        if terraform_check_sum is None:
+            terraform_check_sum = terraform_check_sums.get(version)
+            if terraform_check_sum is None:
+                raise errors.ResourceError(
+                    f"Not stored checksum for version {version} and no checksum provided in spec `terraform_check_sum`")
+        try:
+            download_command = f"wget -c {download_link}"
+            hash_dir = self.getSpace().get_hash_dir()
+            self.execute(download_command, hash_dir)
+            sha256_sum = self.execute(
+                f"sha256sum terraform_{version}_linux_amd64.zip", hash_dir).stdout.decode("UTF-8").split(" ")[0]
+            if sha256_sum != terraform_check_sum:
+                raise errors.ResourceError(
+                    f"downloaded file checksum is {sha256_sum}, does not match {terraform_check_sum}")
+            self.execute(
+                f"unzip -u terraform_{version}_linux_amd64.zip", hash_dir)
+            self.execute(
+                f"mv terraform terraform_{version}_linux_amd64", hash_dir)
+            self.__terraform = os.path.join(
+                hash_dir, f"terraform_{version}_linux_amd64")
+        except Exception as e:
+            raise errors.ResourceError(
+                f"Cannot download terraform binary, error is: {e}")
+
+    def check_terraform(self):
+        version = self.getSpec("terraform_version")
+        if version is None:
+            self.__terraform = "terraform"
+            return
+        if type(version) != str:
+            raise errors.ResourceError(
+                f"terraform_version must be string, found: {type(version)}")
+        self.__terraform = "terraform"
+        if version:
+            try:
+                current_version_output = self.execute(
+                    "terraform version", self.getPath())
+                current_version = current_version_output.stdout.decode(
+                    "UTF-8").split(" ")[1][1:].split("o")[0].strip()
+                if current_version == version:
+                    return
+            except FileNotFoundError as e:
+                pass
+            terraform_install = self.getSpec("terraform_install")
+            if terraform_install is True:
+                self.__install_terraform(version)
+            else:
+                raise errors.ResourceError(
+                    "Terraform version is wrong or not installed and `terraform_install` is not set to true")
+
+    def __get_version(self):
+        try:
+            current_version_output = self.execute(
+                f"{self.__terraform} version", self.getPath())
+            current_version = current_version_output.stdout.decode(
+                "UTF-8").split(" ")[1][1:].split("o")[0].strip()
+            return current_version
+        except FileNotFoundError as e:
+            raise errors.ResourceError(
+                f"No terraform binary found at {self.__terraform}")
+
     @hookimpl
     def action(self, name, state, env):
+        self.check_terraform()
         if name == "build":
             return self.build(state, env)
         elif name == "test":
@@ -489,7 +631,7 @@ class TerraformResource(Resource):
         except subprocess.TimeoutExpired as e:
             raise errors.ResourceBuildError(
                 f"Timeout initializing terraform, {e}")
-        build_command = "terraform plan -out=tfplan -no-color -input=false"
+        build_command = f"{self.__terraform} plan -out=tfplan -no-color -input=false"
         variables = self.getSpec("variables")
         if variables:
             vars = ""
@@ -509,10 +651,10 @@ class TerraformResource(Resource):
             raise errors.ResourceBuildError(e.stderr.decode("UTF-8"))
         artifacts = [self.create_artifact("tfplan", "build", env.getName(), state.get(
             "hash", ""), "file", os.path.join(self.getPath(), "tfplan"))]
-        return {"status": "plan generated", "artifacts": artifacts, "globals": self.get_outputs(self.getPath())}
+        return {"status": "plan generated", "artifacts": artifacts, "globals": self.get_outputs(self.getPath()), "version": self.__get_version()}
 
     def test(self, state: dict, env: EnvResource):
-        fmt_check_command = "terraform fmt --check"
+        fmt_check_command = f"{self.__terraform} fmt --check"
         try:
             self.execute(fmt_check_command, self.getPath())
         except subprocess.CalledProcessError as e:
@@ -523,7 +665,8 @@ class TerraformResource(Resource):
                 f"Timeout checking terraform format, {e}")
         return {
             "error": False,
-            "message": "terraform is fomratted probably"
+            "message": "terraform is fomratted probably",
+            "version": self.__get_version()
         }
 
     def deploy(self, state: dict, env: EnvResource):
@@ -546,7 +689,7 @@ class TerraformResource(Resource):
         except subprocess.TimeoutExpired as e:
             raise errors.ResourceBuildError(
                 f"Timeout initializing terraform, {e}")
-        deploy_command = "terraform apply -input=false -auto-approve -no-color"
+        deploy_command = f"{self.__terraform} apply -input=false -auto-approve -no-color"
         variables = self.getSpec("variables")
         if variables:
             vars = ""
@@ -560,7 +703,7 @@ class TerraformResource(Resource):
         try:
             self.execute(deploy_command, self.getPath(),
                          self.getSpec("apply_timeout", 300))
-            return {"status": "plan applied", "globals": self.get_outputs(self.getPath())}
+            return {"status": "plan applied", "globals": self.get_outputs(self.getPath()), "version": self.__get_version()}
         except FileNotFoundError as e:
             raise errors.ResourceBuildError("cannot find terraform in PATH")
         except subprocess.CalledProcessError as e:
@@ -588,10 +731,29 @@ class KustomizeResource(Resource):
         self.name = self.getMetadata("name")
 
     @classmethod
-    def generate(cls, path, file):
+    def generate(cls, path, file, env: EnvResource, overlay=None, missing_overlay="use_base"):
+        if overlay is None:
+            path_overlay = os.path.join(path, "overlays", env.getName())
+        else:
+            path_overlay = os.path.join(path, "overlays", overlay)
+        if not os.path.exists(path_overlay):
+            if missing_overlay == "use_base":
+                path_overlay = os.path.join(path, "base")
+                if not os.path.exists(path_overlay):
+                    raise errors.ResourceBuildError(
+                        "overlay path and base path not found")
+            elif missing_overlay == "ignore":
+                with open(os.path.join(path, file), "w") as f:
+                    f.write("")
+                return
+            elif missing_overlay == "throw_error":
+                raise errors.ResourceBuildError("overlay path not found")
+            else:
+                raise errors.ResourceBuildError(
+                    f"Unkown value for missing_overlay: {missing_overlay}")
         build_command = "kustomize build"
         try:
-            p = cls.execute(build_command, path)
+            p = cls.execute(build_command, path_overlay)
         except FileNotFoundError as e:
             raise errors.ResourceBuildError("Cannot find kustomize in PATH")
         except subprocess.CalledProcessError as e:
@@ -617,18 +779,66 @@ class KustomizeResource(Resource):
             base_path (str): This path is joined with the path in the spec
             to get the full path to the kustomize files directory.
         """
+        if env is None:
+            raise errors.ResourceBuildError(
+                "Building a kustomize resource requires an env")
         path = os.path.join(
             self.getPath(), f"artifact-manifests-{self.name}.yaml")
-        self.generate(self.getPath(), f"artifact-manifests-{self.name}.yaml")
+        self.generate(self.getPath(), f"artifact-manifests-{self.name}.yaml", env, self.getSpec(
+            "overlay"), self.getSpec("missing_overlay", "use_base"))
+        namespace = self.getSpec("namespace", env.getName())
+        force_namespace = self.getSpec("force_namespace", False)
+        with open(path, "r") as f:
+            manifests = yaml.safe_load_all(f)
+            new_manifests = []
+            for manifest in manifests:
+                if manifest.get("kind") == "Namespace" or manifest.get("kind", "").startswith("Cluster"):
+                    new_manifests.append(manifest)
+                    continue
+                if manifest.get("metadata", {}).get("namespace") is not None and force_namespace:
+                    manifest["metadata"]["namespace"] = namespace
+                elif manifest.get("metadata", {}).get("namespace") is None:
+                    manifest["metadata"]["namespace"] = namespace
+                new_manifests.append(manifest)
+        with open(path, "w") as f:
+            yaml.safe_dump_all(new_manifests, f)
+        env_variables = self.getSpec("envs", {})
+        for name, value in env_variables.items():
+            os.environ[name] = value
+        try:
+            command = f"cat {path} | envsubst"
+            p = self.execute(command, os.path.dirname(path))
+            with open(path, "w") as f:
+                f.write(p.stdout.decode("UTF-8"))
+        except subprocess.CalledProcessError as e:
+            raise errors.ResourceBuildError(
+                f"An error when filling env variables in manifests: {p.stderr.decode('UTF-8')}")
         artifact = self.create_artifact(
             "manifests", "build", env.getName(), state.get("hash", ""), "file", path)
         return {"msg": "manifests generated", "artifacts": [artifact]}
 
     @classmethod
-    def runTests(cls, target: Target, path: str, artifacts: list):
+    def runTests(cls, target: Target, path: str, artifacts: list, env: EnvResource, overlay=None, missing_overlay="use_base"):
+        if overlay is None:
+            path_overlay = os.path.join(path, "overlays", env.getName())
+        else:
+            path_overlay = os.path.join(path, "overlays", overlay)
+        if not os.path.exists(path_overlay):
+            if missing_overlay == "use_base":
+                path_overlay = os.path.join(path, "base")
+                if not os.path.exists(path_overlay):
+                    raise errors.ResourceBuildError(
+                        "overlay path and base path not found")
+            elif missing_overlay == "ignore":
+                return
+            elif missing_overlay == "throw_error":
+                raise errors.ResourceBuildError("overlay path not found")
+            else:
+                raise errors.ResourceBuildError(
+                    f"Unkown value for missing_overlay: {missing_overlay}")
         test_command = "kustomize build -o=/dev/null"
         try:
-            cls.execute(test_command, path)
+            cls.execute(test_command, path_overlay)
         except FileNotFoundError as e:
             raise errors.ResourceTestError("Cannot find kustomize in PATH")
         except subprocess.CalledProcessError as e:
@@ -665,7 +875,8 @@ class KustomizeResource(Resource):
 
     def test(self, state: dict, env: EnvResource):
         if env is None:
-            artifacts = state["no_env"]["build"].get("artifacts", [])
+            raise errors.ResourceTestError(
+                "Testing a kustomize resource requires an env")
         else:
             artifacts = state["envs"][env.getName()]["build"].get(
                 "artifacts", [])
@@ -673,12 +884,20 @@ class KustomizeResource(Resource):
             if target is None:
                 raise errors.ResourceTestError(
                     f"No target of kind K8STarget in env {env}")
-        self.runTests(target, self.getPath(), artifacts)
+        namespace = self.getSpec("namespace", env.getName())
+        try:
+            self.execute(
+                f"kubectl create namespace {namespace}", self.getPath())
+        except subprocess.CalledProcessError:
+            pass
+        self.runTests(target, self.getPath(), artifacts, env, self.getSpec(
+            "overlay"), self.getSpec("missing_overlay", "use_base"))
         return {"msg": "kustomize tested without errors"}
 
     def deploy(self, state: dict, env: EnvResource):
         if env is None:
-            artifacts = state["no_env"]["build"].get("artifacts", [])
+            raise errors.ResourceTestError(
+                "Deploying a kustomize resource requires an env")
         else:
             artifacts = state["envs"][env.getName()]["build"].get(
                 "artifacts", [])
@@ -846,7 +1065,7 @@ class GoServiceResource(Resource):
         self.name = self.getMetadata("name")
 
     @ classmethod
-    def compile(cls, go_executable, executable_name, path):
+    def compile(cls, go_executable, executable_name, path, go_mod_path: str):
         go_version_command = f"{go_executable} version"
         ret = {}
         try:
@@ -855,12 +1074,18 @@ class GoServiceResource(Resource):
             ret["version"] = r.split(" ")[2][2:]
         except subprocess.CalledProcessError as e:
             raise errors.ResourceBuildError(e)
+        go_mod_tidy_command = f"{go_executable} mod tidy"
+        try:
+            cls.execute(go_mod_tidy_command, go_mod_path)
+        except subprocess.CalledProcessError as e:
+            raise errors.ResourceBuildError(e.stderr.decode("UTF-8"))
         compile_command = f"{go_executable} build -o {executable_name} ."
         try:
+            os.environ["CGO_ENABLED"] = "0"
             cls.execute(compile_command, path)
             return ret
         except subprocess.CalledProcessError as e:
-            raise errors.ResourceBuildError(e)
+            raise errors.ResourceBuildError(e.stderr.decode("UTF-8"))
 
     @ classmethod
     def runTests(cls, go_executable, golangci_lint_executable, path):
@@ -890,7 +1115,7 @@ class GoServiceResource(Resource):
 
     def build(self, state: dict, env: EnvResource):
         executable_name = self.getSpec("exec_name", self.name)
-        self.compile("go", executable_name, self.getPath())
+        self.compile("go", executable_name, self.getPath(), self.getPath())
         artifact = self.create_artifact("binary", "build", env.getName(
         ), state["hash"], "file", os.path.join(self.getPath(), executable_name))
         return {
@@ -911,7 +1136,7 @@ class GoServiceResource(Resource):
 
 class MicroService(object):
     @ classmethod
-    def build(cls, res, state, target):
+    def build(cls, res, state, target, env: EnvResource):
         docker_path = os.path.join(
             res.getPath(), res.getSpec("docker_path", "."))
         image_name = DockerImageResource.build_image_name(res, state)
@@ -921,7 +1146,7 @@ class MicroService(object):
         }
         image_url = target.action("publish", config)
         k8s_path = os.path.join(res.getPath(), res.getSpec("k8s_path", "k8s"))
-        KustomizeResource.generate(k8s_path, "manifests.yaml")
+        KustomizeResource.generate(k8s_path, "manifests.yaml", env)
         k8s_file = os.path.join(k8s_path, "manifests.yaml")
         try:
             with open(k8s_file, "r") as f:
@@ -948,7 +1173,7 @@ class MicroService(object):
         return image_url, k8s_file
 
     @ classmethod
-    def test(cls, res, target, artifacts):
+    def test(cls, res, target, artifacts, env: EnvResource):
         # Test docker image
         image_name = res.getSpec("image_name", res.name)
         DockerImageResource.runTests(image_name, res.getPath())
@@ -958,7 +1183,7 @@ class MicroService(object):
             if artifact.getId() == "k8s":
                 k8s_path = os.path.join(
                     res.getPath(), res.getSpec("k8s_path", "k8s"))
-                KustomizeResource.runTests(target, k8s_path, [artifact])
+                KustomizeResource.runTests(target, k8s_path, [artifact], env)
                 break
 
     @ classmethod
@@ -989,6 +1214,8 @@ class GoMicroServiceResource(Resource):
                 "This is not a GoMicroService resource")
         self.name = self.getMetadata("name")
         self.__go_checksums = {
+            "1.20.1": "000a5b1fca4f75895f78befeb2eecf10bfff3c428597f3f1e69133b63b911b02",
+            "1.20": "5a9ebcc65c1cce56e0d2dc616aff4c4cedcfbda8cc6f0288cc08cda3b18dcbf1",
             "1.19.3": "74b9640724fd4e6bb0ed2a1bc44ae813a03f1e72a4c76253e2d5c015494430ba",
             "1.19.2": "5e8c5a74fe6470dd7e055a461acda8bb4050ead8c2df70f227e3ff7d8eb7eeb6",
             "1.19.1": "acc512fbab4f716a8f97a8b3fbaa9ddd39606a28be6c2515ef7c6c6311acffde",
@@ -1004,6 +1231,8 @@ class GoMicroServiceResource(Resource):
             "1.18.8": "4d854c7bad52d53470cf32f1b287a5c0c441dc6b98306dea27358e099698142a",
         }
         self.__golangci_lint_checksums = {
+            "1.51.1": "17aeb26c76820c22efa0e1838b0ab93e90cfedef43fbfc9a2f33f27eb9e5e070",
+            "1.51.0": "38c25ae0ba5bfebd3ec42e9230547bd6461b179e47d7ba4d86950623bf28862a",
             "1.50.1": "4ba1dc9dbdf05b7bdc6f0e04bdfe6f63aa70576f51817be1b2540bbce017b69a",
             "1.50.0": "b4b329efcd913082c87d0e9606711ecb57415b5e6ddf233fde9e76c69d9b4e8b",
             "1.49.0": "5badc6e9fee2003621efa07e385910d9a88c89b38f6c35aded153193c5125178"
@@ -1013,6 +1242,9 @@ class GoMicroServiceResource(Resource):
 
     def check_go(self):
         go_version = self.getSpec("go_version")
+        if type(go_version) != str:
+            raise errors.ResourceError(
+                f"go_version must be string found: {type(go_version)}")
         if go_version:
             try:
                 if os.path.isfile(os.path.join(self.getSpace().get_hash_dir(), f"go{go_version}.linux-amd64", "bin", "go")):
@@ -1026,6 +1258,9 @@ class GoMicroServiceResource(Resource):
 
     def check_golangci_lint(self):
         check_golangci_lint_version = self.getSpec("golangci_lint_version")
+        if type(check_golangci_lint_version) != str:
+            raise errors.ResourceError(
+                f"golangci_lint_version must be string found: {type(check_golangci_lint_version)}")
         if check_golangci_lint_version:
             try:
                 if os.path.isfile(os.path.join(self.getSpace().get_hash_dir(), f"golangci-lint-{check_golangci_lint_version}-linux-amd64", "golangci-lint")):
@@ -1113,6 +1348,10 @@ class GoMicroServiceResource(Resource):
             except subprocess.CalledProcessError as e:
                 raise errors.ActionError(e)
 
+    @hookimpl
+    def get_ignores(self):
+        return [".deployed"]
+
     @ hookimpl
     def action(self, name, state, env):
         self.setup_tools()
@@ -1127,42 +1366,83 @@ class GoMicroServiceResource(Resource):
         if env is None:
             raise errors.ResourceBuildError(
                 f"Building micro service {self} requires an env")
-        target = env.getTarget("DockerRegistryTarget")
-        if target is None:
+        deploy_target = self.getSpec("deploy_target", "kubernetes")
+        if deploy_target == "kubernetes":
+            target = env.getTarget("DockerRegistryTarget")
+            if target is None:
+                raise errors.ResourceBuildError(
+                    f"No target of kind DockerRegistry in env {env}")
+        elif deploy_target == "DOFunction":
+            target = env.getTarget("DOFunctionTarget")
+            if target is None:
+                raise errors.ResourceBuildError(
+                    f"No target of kind DOFunction in env {env}")
+        else:
             raise errors.ResourceBuildError(
-                f"No target of kind DockerRegistry in env {env}")
+                f"Invalid value for 'deploy_target': {deploy_target}, valid values are: 'kubernetes' and 'DOFunction'")
         # Build go service
         executable_name = self.getSpec("exec_name", self.name)
+        go_mod_path = self.getSpec("go_mod_path")
+        if go_mod_path is None:
+            go_mod_path = self.getSpace().get_base()
+        else:
+            go_mod_path = os.path.join(
+                self.getPath(), go_mod_path)
+        h = core.Hash()
+        go_mod_hash = h.hash(go_mod_path, ["go.mod", "go.sum"])
         go_env_data = GoServiceResource.compile(
-            self.__go, executable_name, self.getPath())
+            self.__go, executable_name, self.getPath(), go_mod_path)
         executable_path = os.path.join(self.getPath(), executable_name)
-        image_url, k8s_file = MicroService.build(self, state, target)
 
         binary_artifact = self.create_artifact(
             "binary", "build", env.getName(), state["hash"], "file", executable_path)
+        if deploy_target == "DOFunction":
+            return {
+                "artifacts": [binary_artifact],
+                "_go": go_env_data,
+                "_go_mod_hash": go_mod_hash
+            }
+        image_url, k8s_file = MicroService.build(self, state, target, env)
         k8s_artifact = self.create_artifact(
             "k8s", "build", env.getName(), state["hash"], "file", k8s_file)
         image_url_artifact = self.create_artifact(
             "image_url", "build", env.getName(), state["hash"], "url", image_url)
         return {
             "artifacts": [binary_artifact, k8s_artifact, image_url_artifact],
-            "_go": go_env_data
+            "_go": go_env_data,
+            "_go_mod_hash": go_mod_hash
         }
 
     def test(self, state: dict, env: EnvResource):
         if env is None:
             raise errors.ResourceTestError(
                 f"Testing micro service {self} requires an env")
-        target = env.getTarget("K8STarget")
-        if target is None:
+        deploy_target = self.getSpec("deploy_target", "kubernetes")
+        if deploy_target == "kubernetes":
+            target = env.getTarget("K8STarget")
+            if target is None:
+                raise errors.ResourceTestError(
+                    f"No target of kind K8STarget in env {env}")
+        elif deploy_target == "DOFunction":
+            target = env.getTarget("DOFunctionTarget")
+            if target is None:
+                raise errors.ResourceTestError(
+                    f"No target of kind DOFunction in env {env}")
+        else:
             raise errors.ResourceTestError(
-                f"No target of kind K8STarget in env {env}")
+                f"Invalid value for 'deploy_target': {deploy_target}, valid values are: 'kubernetes' and 'DOFunction'")
         artifacts = state["envs"][env.getName()]["build"].get("artifacts", [])
         # Test go code
         golang_ci_lint_env_data = GoServiceResource.runTests(
             self.__go, self.__golangci_lint, self.getPath())
-
-        MicroService.test(self, target, artifacts)
+        if deploy_target == "kubernetes":
+            namespace = self.getSpec("namespace", env.getName())
+            try:
+                self.execute(
+                    f"kubectl create namespace {namespace}", self.getPath())
+            except subprocess.CalledProcessError:
+                pass
+            MicroService.test(self, target, artifacts, env)
         return {
             "_golangci_lint": golang_ci_lint_env_data
         }
@@ -1172,19 +1452,68 @@ class GoMicroServiceResource(Resource):
         if env is None:
             raise errors.ResourceDeployError(
                 f"Deploying micro service {self} requires an env")
-        target = env.getTarget("K8STarget")
-        if target is None:
-            raise errors.ResourceDeployError(
-                f"No target of kind K8STarget in env {env}")
+        deploy_target = self.getSpec("deploy_target", "kubernetes")
+        if deploy_target == "kubernetes":
+            target = env.getTarget("K8STarget")
+            if target is None:
+                raise errors.ResourceTestError(
+                    f"No target of kind K8STarget in env {env}")
+        elif deploy_target == "DOFunction":
+            target = env.getTarget("DOFunctionTarget")
+            if target is None:
+                raise errors.ResourceTestError(
+                    f"No target of kind DOFunction in env {env}")
+        else:
+            raise errors.ResourceTestError(
+                f"Invalid value for 'deploy_target': {deploy_target}, valid values are: 'kubernetes' and 'DOFunction'")
         artifacts = state["envs"][env.getName()]["build"].get("artifacts", [])
-        MicroService.deploy(self, target, artifacts)
+        if deploy_target == "kubernetes":
+            MicroService.deploy(self, target, artifacts)
+        else:
+            url = target.action("deploy", {
+                "project_dir": self.getPath(),
+                "package_name": self.getSpec("do_function", {}).get("package_name", self.getName()),
+                "function_name": self.getSpec("do_function", {}).get("function_name", self.getName())
+            })
+            return {
+                "globals": {
+                    "do_function_url": url
+                }
+            }
 
     @ hookimpl
-    def re_action(self, action, state, env):
+    def re_action(self, action, state, env: EnvResource):
         if env is None:
             result = state.get("no_env")
         else:
             result = state.get("envs", {}).get(env.getName())
+        go_mod_path = self.getSpec("go_mod_path")
+        if go_mod_path is None:
+            go_mod_path = self.getSpace().get_base()
+        else:
+            go_mod_path = os.path.join(self.getPath(), go_mod_path)
+        h = core.Hash().hash(go_mod_path, ["go.mod", "go.sum"])
+        go_mod_hash = result.get("build", {}).get("_go_mod_hash")
+        if go_mod_hash != h:
+            return True
+        if action == "deploy" and env is not None:
+            target = env.getTarget("K8STarget")
+            if target:
+                artifacts = state.get("envs", {}).get(
+                    env.getName(), {}).get("build", {}).get("artifacts", [])
+                for artifact in artifacts:
+                    if artifact.getId() != "k8s":
+                        continue
+                    config = {
+                        "path": self.getPath(),
+                        "manifests_path": artifact.getData(),
+                        "diff": True
+                    }
+                    try:
+                        return target.action("test", config)
+                    except errors.ResourceError as e:
+                        raise errors.ResourceTestError(str(e))
+
         self.check_go()
         go_version = result.get("build", {}).get("_go", {}).get("version")
         go_version_command = f"{self.__go} version"
@@ -1344,4 +1673,5 @@ class ResourceSpace(object):
         if resource.getKind() == "Env":
             match = [resource.getFile().split("/")[-1]]
             match.extend(resource.getSpec("match", []))
+        ignore.extend(resource.get_ignores())
         return core.Hash(alg).hash(os.path.join(self.__base, resource.getPath()), match, ignore)
