@@ -1,5 +1,6 @@
 import os
 import json
+import yaml
 from subprocess import CalledProcessError
 import tempfile
 import pluggy
@@ -227,8 +228,10 @@ class K8STarget(Target):
                     f"No manifests path supplied for K8S target")
             if name == "deploy":
                 kubectl_command = f"kubectl apply {kubectl_args} -f {manifests_path}"
-            elif name == "test":
+            elif name == "test" and config.get("diff") is not True:
                 kubectl_command = f"kubectl apply {kubectl_args} --dry-run=server -f {manifests_path}"
+            elif name == "test" and config.get("diff") is True:
+                kubectl_command = f"kubectl diff {kubectl_args} -f {manifests_path}"
         elif config_kind == "DockerImage":
             image_url = config.get("image_url")
             if image_url is None:
@@ -239,6 +242,8 @@ class K8STarget(Target):
         try:
             resources.Resource.execute(kubectl_command, path)
         except CalledProcessError as e:
+            if config.get("diff") is True and e.returncode == 1:
+                return True
             raise errors.ResourcePublishError(e)
         finally:
             if k8s_certificate_ca:
@@ -354,6 +359,102 @@ class DockerRegistryTarget(Target):
                 raise errors.ResourcePublishError(e)
 
 
+class DOFunctionTarget(Target):
+
+    def __verify_function(self, package_name: str, function_name: str, config: dict) -> bool:
+        packages = config.get("packages", [])
+        for package in packages:
+            if package.get("name") == package_name:
+                functions = package.get("functions", [])
+                for function in functions:
+                    if function.get("name") == function_name:
+                        return True
+        return False
+
+    @hookimpl
+    def init(self, config: dict, space) -> None:
+        self._init(config)
+        if self.kind != "DOFunctionTarget":
+            raise errors.TargetConfigError(
+                f"This target is not of kind DOFunction, it is of kind {self.kind}")
+        self.__space = space
+
+    @hookimpl
+    def action(self, name: str, config: dict):
+        if name != "deploy":
+            raise errors.ResourceError(
+                f"DOFunction target can only run deploy actions not {name}")
+        region = self.spec.get("region")
+        if region is None:
+            raise errors.ResourceError(
+                "No region found in DOFunction target specs")
+        project_dir = config.get('project_dir')
+        if project_dir is None:
+            raise errors.ResourceError(
+                "No project_dir provided in config for DOFunction target")
+        if not os.path.exists(os.path.join(project_dir, 'project.yml')):
+            raise errors.ResourceError(
+                f"No project.yml file found in {project_dir}")
+        package_name = config.get("package_name")
+        if package_name is None:
+            raise errors.ResourceError(
+                "No package_name is provided for DOFunction target")
+        function_name = config.get("function_name")
+        if function_name is None:
+            raise errors.ResourceError(
+                "No function_name is provided for DOFunction target")
+        with open(os.path.join(project_dir, 'project.yml'), "r") as f:
+            project_yaml = yaml.safe_load(f)
+        if not self.__verify_function(package_name, function_name, project_yaml):
+            raise errors.ResourceError(
+                f"No package with name {package_name} and fuction with {function_name} found in {os.path.join(project_dir, 'project.yml')}")
+        function_index = f"{package_name}/{function_name}"
+        label = self.spec.get("label")
+        if label is None:
+            raise errors.ResourceError(
+                "No label found in DOFunction target specs")
+        namespaces_list_command = "doctl serverless namespaces list -ojson"
+        namespace_id = None
+        try:
+            p = resources.Resource.execute(
+                namespaces_list_command, project_dir)
+            namespaces = p.stdout.decode("UTF-8")
+            namespaces_list = json.loads(namespaces)
+            for namespace in namespaces_list:
+                if namespace.get("label") == label:
+                    namespace_id = namespace.get("namespace")
+        except CalledProcessError as e:
+            raise errors.ResourceError(e.stdout.decode("UTF-8"))
+        if namespace_id is None:
+            create_namespace_command = f"doctl serverless namespaces create --label {label} --region {region}"
+            try:
+                p = resources.Resource.execute(
+                    create_namespace_command, project_dir)
+                namespace_id = p.stdout.decode("UTF-8").split(' ')[4]
+            except CalledProcessError as e:
+                raise errors.ResourceError(
+                    f"Error creating namespace in DO: {e.stderr.decode('UTF-8')}")
+        try:
+            resources.Resource.execute(
+                f"doctl serverless connect {namespace_id}", project_dir)
+        except CalledProcessError as e:
+            raise errors.ResourceError(
+                f"Error connecting to namespace: {e.stderr.decode('UTF-8')}")
+        deploy_command = f"doctl serverless deploy . -ojson"
+        try:
+            resources.Resource.execute(
+                deploy_command, config.get('project_dir'))
+        except CalledProcessError as e:
+            raise errors.ResourceError(e.stdout.decode("UTF-8"))
+        try:
+            p = resources.Resource.execute(
+                f"doctl sls fn get {function_index} --url", project_dir)
+            return p.stdout.decode("UTF-8")
+        except CalledProcessError as e:
+            raise errors.ResourceError(
+                f"Error when getting function URL: {function_index}, {e.stderr.decode('UTF-8')}")
+
+
 def get_plugin_manager():
     """
     Return the plugin manager for hash-targets plugins.
@@ -364,6 +465,7 @@ def get_plugin_manager():
     pm.register(FakeTarget(), "FakeTarget")
     pm.register(K8STarget(), "K8STarget")
     pm.register(DockerRegistryTarget(), "DockerRegistryTarget")
+    pm.register(DOFunctionTarget(), "DOFunctionTarget")
     return pm
 
 
