@@ -134,6 +134,8 @@ class HashStore(object):
                                 saved=True
                             )
                             f.write(self.read_artifact(aft, resource_id))
+                            os.chmod(artf.get("data"),
+                                     artf.get("file_mode", 33204))
         for ac in core.all_actions:
             artfs = state.get("no_env", {}).get(
                 ac, {}).get("artifacts", [])
@@ -150,6 +152,7 @@ class HashStore(object):
                             saved=True
                         )
                         f.write(self.read_artifact(aft, resource_id))
+                        os.chmod(art.get("data"), art.get("file_mode"))
 
     def get(self, resource_id: str, h: str):
         """
